@@ -3,13 +3,13 @@ from __future__ import annotations
 import abc
 import asyncio
 from collections import defaultdict
-import functools
 import hashlib
 import logging
 import secrets
 import socket
 from typing import (
     Any,
+    Awaitable,
     Callable,
     ClassVar,
     Coroutine,
@@ -25,13 +25,13 @@ from typing import (
     cast,
 )
 import uuid
-import weakref
 
 import aioredis
 import aioredis.exceptions
 import aioredis.sentinel
 from aiotools.context import aclosing
 from aiotools.server import process_index
+from aiotools.ptaskgroup import PersistentTaskGroup
 import attr
 
 from . import msgpack, redis
@@ -599,46 +599,47 @@ class EventDispatcher(aobject):
     redis_client: RedisConnectionInfo
     consumer_loop_task: asyncio.Task
     subscriber_loop_task: asyncio.Task
-    consumer_taskset: weakref.WeakSet[asyncio.Task]
-    subscriber_taskset: weakref.WeakSet[asyncio.Task]
+    consumer_taskgroup: PersistentTaskGroup
+    subscriber_taskgroup: PersistentTaskGroup
 
     _log_events: bool
     _consumer_name: str
 
     def __init__(
         self,
-        connector: EtcdRedisConfig,
+        redis_config: EtcdRedisConfig,
         db: int = 0,
         log_events: bool = False,
         *,
         consumer_group: str = "manager",
         node_id: str = None,
+        consumer_exception_handler: Callable[[Exception], Awaitable[None]] = None,
+        subscriber_exception_handler: Callable[[Exception], Awaitable[None]] = None,
     ) -> None:
-        self.redis_client = redis.get_redis_object(connector, db=db)
+        self.redis_client = redis.get_redis_object(redis_config, db=db)
         self._log_events = log_events
         self.consumers = defaultdict(set)
         self.subscribers = defaultdict(set)
         self._consumer_group = consumer_group
         self._consumer_name = _generate_consumer_id(node_id)
+        self.consumer_taskgroup = PersistentTaskGroup(
+            name="consumer_taskgroup",
+            exception_handler=consumer_exception_handler,
+        )
+        self.subscriber_taskgroup = PersistentTaskGroup(
+            name="subscriber_taskgroup",
+            exception_handler=subscriber_exception_handler,
+        )
 
     async def __ainit__(self) -> None:
         self.consumer_loop_task = asyncio.create_task(self._consume_loop())
         self.subscriber_loop_task = asyncio.create_task(self._subscribe_loop())
-        self.consumer_taskset = weakref.WeakSet()
-        self.subscriber_taskset = weakref.WeakSet()
 
     async def close(self) -> None:
         try:
             cancelled_tasks = []
-            for task in self.consumer_taskset:
-                if not task.done():
-                    task.cancel()
-                    cancelled_tasks.append(task)
-            for task in self.subscriber_taskset:
-                if not task.done():
-                    task.cancel()
-                    cancelled_tasks.append(task)
-            await asyncio.sleep(0)
+            await self.consumer_taskgroup.shutdown()
+            await self.subscriber_taskgroup.shutdown()
             self.consumer_loop_task.cancel()
             self.subscriber_loop_task.cancel()
             cancelled_tasks.append(self.consumer_loop_task)
@@ -692,7 +693,6 @@ class EventDispatcher(aobject):
         self.subscribers[handler.event_cls.name].discard(cast(EventHandler[Any, AbstractEvent], handler))
 
     async def handle(self, evh_type: str, evh: EventHandler, source: AgentId, args: tuple) -> None:
-        loop = asyncio.get_running_loop()
         coalescing_opts = evh.coalescing_opts
         coalescing_state = evh.coalescing_state
         cb = evh.callback
@@ -704,10 +704,7 @@ class EventDispatcher(aobject):
                 # mypy cannot catch the meaning of asyncio.iscoroutinefunction().
                 await cb(evh.context, source, event_cls.deserialize(args))  # type: ignore
             else:
-                wrapped_cb = functools.partial(
-                    cb, evh.context, source, event_cls.deserialize(args),  # type: ignore
-                )
-                loop.call_soon(wrapped_cb)
+                cb(evh.context, source, event_cls.deserialize(args))  # type: ignore
 
     async def dispatch_consumers(
         self,
@@ -718,9 +715,9 @@ class EventDispatcher(aobject):
         if self._log_events:
             log.debug('DISPATCH_CONSUMERS(ev:{}, ag:{})', event_name, source)
         for consumer in self.consumers[event_name].copy():
-            self.consumer_taskset.add(asyncio.create_task(
-                self.handle("CONSUMER", consumer, source, args)
-            ))
+            self.consumer_taskgroup.create_task(
+                self.handle("CONSUMER", consumer, source, args),
+            )
             await asyncio.sleep(0)
 
     async def dispatch_subscribers(
@@ -732,9 +729,9 @@ class EventDispatcher(aobject):
         if self._log_events:
             log.debug('DISPATCH_SUBSCRIBERS(ev:{}, ag:{})', event_name, source)
         for subscriber in self.subscribers[event_name].copy():
-            self.subscriber_taskset.add(asyncio.create_task(
-                self.handle("SUBSCRIBER", subscriber, source, args)
-            ))
+            self.subscriber_taskgroup.create_task(
+                self.handle("SUBSCRIBER", subscriber, source, args),
+            )
             await asyncio.sleep(0)
 
     async def _consume_loop(self) -> None:
@@ -745,6 +742,8 @@ class EventDispatcher(aobject):
             self._consumer_name,
         )) as agen:
             async for msg_id, msg_data in agen:
+                if msg_data is None:
+                    continue
                 try:
                     await self.dispatch_consumers(
                         msg_data[b'name'].decode(),
@@ -762,6 +761,8 @@ class EventDispatcher(aobject):
             'events',
         )) as agen:
             async for msg_id, msg_data in agen:
+                if msg_data is None:
+                    continue
                 try:
                     await self.dispatch_subscribers(
                         msg_data[b'name'].decode(),
