@@ -14,6 +14,7 @@ from typing import (
     Optional,
     Sequence,
     Tuple,
+    Union,
 )
 
 import aioredis
@@ -257,7 +258,7 @@ async def execute(
             await asyncio.sleep(reconnect_poll_interval)
             continue
         except aioredis.exceptions.ResponseError as e:
-            if e.args[0].startswith("NOREPLICAS "):
+            if "NOREPLICAS" in e.args[0]:
                 await asyncio.sleep(reconnect_poll_interval)
                 continue
             raise
@@ -274,7 +275,7 @@ async def execute_script(
     script_id: str,
     script: str,
     keys: Sequence[str],
-    args: Sequence[str],
+    args: Sequence[Union[bytes, memoryview, str, int, float]],  # aioredis.connection.EncodableT
 ) -> Any:
     """
     Auto-load and execute the given script.
@@ -322,7 +323,7 @@ async def read_stream(
     """
     A high-level wrapper for the XREAD command.
     """
-    last_id = b'0-0'
+    last_id = b'$'
     while True:
         try:
             reply = await execute(
@@ -334,6 +335,16 @@ async def read_stream(
             )
             if reply is None:
                 continue
+            # Keep some latest messages so that other manager
+            # processes to have chances of fetching them.
+            await execute(
+                r,
+                lambda r: r.xtrim(
+                    stream_key,
+                    maxlen=128,
+                    approximate=True,
+                ),
+            )
             for msg_id, msg_data in reply[0][1]:
                 try:
                     yield msg_id, msg_data
@@ -356,7 +367,6 @@ async def read_stream_by_group(
     A high-level wrapper for the XREADGROUP command
     combined with XAUTOCLAIM and XGROUP_CREATE.
     """
-    last_ack = b"0-0"
     while True:
         try:
             messages = []
@@ -390,20 +400,16 @@ async def read_stream_by_group(
             assert reply[0][0].decode() == stream_key
             for msg_id, msg_data in reply[0][1]:
                 messages.append((msg_id, msg_data))
+            await execute(
+                r,
+                lambda r: r.xack(
+                    stream_key,
+                    group_name,
+                    *(msg_id for msg_id, msg_data in reply[0][1]),
+                ),
+            )
             for msg_id, msg_data in messages:
-                try:
-                    yield msg_id, msg_data
-                finally:
-                    if _parse_stream_msg_id(last_ack) < _parse_stream_msg_id(msg_id):
-                        last_ack = msg_id
-                    await execute(
-                        r,
-                        lambda r: r.xack(
-                            stream_key,
-                            group_name,
-                            msg_id,
-                        ),
-                    )
+                yield msg_id, msg_data
         except asyncio.CancelledError:
             raise
         except aioredis.exceptions.ResponseError as e:
